# Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
This module contains code that is common to all face recognition algorithms
and defines a interfaces that can be adopted.

Created on Aug 16, 2010

@author: bolme
'''

import pyvision as pv
import numpy as np

class FaceRecognitionAlgorithm:
    '''
    This class defines an interface to a face recogintion algorithm.
    Training is expected to be unique to each algorithm.  After the
    algorithms are trained instances that implement this interface
    could be easily substituded for each other with little code 
    modification.
    '''
    
    def addTraining(self,label,im,rect,leye,reye,ilog=None):
        '''
        Add a training face.  This is a suggested interface that should 
        work for many algorithms and therefore can use existing training
        code.
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        raise NotImplementedError()
    
    
    
    def addCohort(self,im,rect,leye,reye,ilog=None):
        '''
        Add a face to use for cohort normalization.  This is a suggested 
        interface that should work for many algorithms.  Unless superseded 
        in subclasses that use cohort normalization this method results
        in a no-op.
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        pass # NO OP
    
    
    
    def train(self,ilog=None):
        '''
        Train the algorithm.  After all training faces are added this 
        function will be called to do additional computation and processing
        

        @param ilog: optional image log that is used to save intermediate
                        training data such as images of the eigenvectors.
        @type ilog: pv.ImageLog
        '''
        raise NotImplementedError()
    
    def getFaceRecord(self,im,rect,leye,reye,ilog=None):
        '''
        This function computes a face record.  It takes an image of an 
        unknown face and produces a low dimensional represetation that 
        is used for storage and matching.  The original image can then
        be freed.  For example, in the eigen faces algoirthm this function
        would preprocess the image and then project it onto the eigenbasis.
        What is returned is a face record that contains data used to match
        the face along with other metadata for that image.        
                
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (Not Used)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: an ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        
        @return: A face record object
        @rtype: FaceRecord
        '''
        raise NotImplementedError()
    
    
    def similarity(self,face_record1,face_record2):
        '''
        A similarity score computed for two face records.  High values 
        indicate a better match.
        
        @param face_record1: a face record created by calling getFaceRecord.
        @type face_record1: FaceRecord
        
        @param face_record2: a face record created by calling getFaceRecord.
        @type face_record2: FaceRecord
        
        @returns: the similarity as a floating point value
        @rtype: float
        '''
        raise NotImplementedError()
    
    
    def similarityMatrix(self,probes,targets):
        '''
        A similarity matrix computed between sets of face records.  This 
        could be implemented by multiple calls to the "similarity" method, 
        but can be overridden to achieve faster performance when computing
        multiple comparisons.
        
        @param probes: a list of face records.
        @type probes: [FaceRecord, ...]
        
        @param targets: a list of face records.
        @type targets: [FaceRecord, ...]
        
        @return: a similarity matrix in numpy format.
        @rtype: numpy.array
        '''
        raise NotImplementedError()




    
    
    
class FaceRecord:
    '''This is a simple structure to organize information on a face.'''
    
    def __init__(self, face_rect, left_eye, right_eye):
        self.face_rect = face_rect
        self.left_eye  = left_eye
        self.right_eye = right_eye

    
    
def colorMix(tile,weights,hue_adj):
    '''
    This function splits the image into seperate color channels: RGB, HSV, LAB.
    It then remixes the image based on the weights.  In theory, some remixings
    should be better for recogintion.  For example, the red channel seems to work
    well for LDA based approaches.
    '''
    # Compute color spaces
    rgb = tile.asMatrix3D()
    hsv = pv.Image(tile.asHSV()).asMatrix3D()
    lab = pv.Image(tile.asLAB()).asMatrix3D()
    
    # Break out channels
    red = rgb[0,:,:]
    green = rgb[1,:,:]
    blue = rgb[2,:,:]
    v = hsv[0,:,:]
    s = hsv[1,:,:]
    h = hsv[2,:,:]
    b = lab[0,:,:]
    a = lab[1,:,:]
    l = lab[2,:,:]
    h = 2*np.pi*h/180.0
    h1 = 128*np.sin(h+hue_adj)
    h2 = 128*np.cos(h+hue_adj)
    a = a - 128.0
    b = b - 128.0
    
    colors = [red,green,blue,h1,h2,s,l,a,b]
    
    # remix channels
    assert len(colors) == 9
    assert len(weights) == 9
    colors = np.array(colors)
    color_mix = np.array(weights).reshape(9,1,1)
    tile = (colors*color_mix).sum(axis=0)
    
    # return the tile in the remixed color space
    return tile


    
    
    
    
    
    
    