# Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Created on Jan 29, 2011

@author: bolme
'''
import common
import pyvision as pv
import numpy as np
import time

GBU_TUNING = {
              'pca_keep': 1133, 
              'eye_y': 0.42579601645217319, 
              'sigma_low': 17.249069969341434, 
              'eye_width': 0.4476571056046178, 
              'color': [-0.9593161302209543, 0.98112965971539867, 0, -0.025001331645825998, 0.037695652569289008, -0.048151653059113578, -0.25028374642441487, -0.79943048584179, -0.13577037554439905-0.54921551816773395], 
              'sigma_high': 1.6106343852657956, 
              'mean_std': True, 
              'hue_adj': 3.798525610894806, 
              'sim': 'CORR', 
              'tile_size': [76, 100], 
              'band_pass': False, 
              'reg': 0.030891031296346461, 
              'log_transform': False, 
              'log_shift': 1.0,
              'lda_keep': 297,
              'feature_norm': False
              }


#use with CohortLDA
CohortLDA_REGIONS = [
                    { # Red channel
                      'tile_size': (65,75), 
                      'eye_y': None, 
                      'eye_width': None, 
                      'rect': pv.Rect(0,0,1,1),
                      'log_transform': True, 
                      'log_shift': 0.1,
                      'color': [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], 
                      'sigma_low': None, 
                      'sigma_high': None, 
                      'hue_adj': 0.0, 
                      'mean_std': True, 
                      'pca_keep': 0.98,
                      'cohort_match': None,
                      'feature_norm':True,
                      'reg': 2.0**-6, 
                      'lda_keep': 128,
                      'sim': 'L2',
                    },
                    { # I channel from YIQ
                      'tile_size': (65,75), 
                      'eye_y': None, 
                      'eye_width': None, 
                      'rect': pv.Rect(0,0,1,1),
                      'log_transform': False, 
                      'log_shift': 0.1,
                      'color': [0.595716,-0.274453,-0.321263,0.0,0.0,0.0,0.0,0.0,0.0], 
                      'sigma_low': None, 
                      'sigma_high': None, 
                      'hue_adj': 0.0, 
                      'mean_std': True, 
                      'pca_keep': 0.98,
                      'cohort_match': None,
                      'feature_norm':True,
                      'reg': 2.0**-6, 
                      'lda_keep': 128,
                      'sim': 'L2',
                   },
                ]
'''Tuning for the CohortLDA baseline algorithms.  Used with the GBU_Uncontrolled_x8 dataset.'''
CohortLDA_KEYWORDS = {         
                'tile_size': [65, 75], 
                'eye_y': 0.33333333333, 
                'eye_width': 0.4230769230769231, 
                'smoothing':0.5,
}
'''Tuning for the CohortLDA baseline algorithms.  Used with the GBU_Uncontrolled_x8 dataset.'''


class LDA(common.FaceRecognitionAlgorithm):
    '''
    This class defines an interface to a face recogintion algorithm.
    Training is expected to be unique to each algorithm.  After the
    algorithms are trained instances that implement this interface
    could be easily substituded for each other with little code 
    modification.
    '''
    def __init__(self,
                 eye_y = 0.35,
                 eye_width = 0.55,
                 tile_size = (65,85),
                 color = [0.299,0.587,0.114,0,0,0,0,0,0],
                 hue_adj = 0.0,
                 pca_keep = 1500,
                 lda_keep = 125,
                 log_transform = True,
                 log_shift = 1.0,
                 mean_std = True,
                 sigma_high = 0.5,
                 sigma_low = 5.0, 
                 sim = "L2",
                 reg = 0.1,
                 sub_rect = None,
                 cohort_adjust=True,
                 cohort_match=None,
                 feature_norm=False,
                 smoothing = None,
                 ):
        '''
        Initialize an LDA face recognition algorithm.

        @param eye_y:          The y location of the eyes relative to the height of 
                               the geometrically normalized image.  Set to None if the
                               tile is already geometrically transformed.
        @type eye_y:           float or None
        @param eye_width:      The width location of the eyes relative to the width of 
                               the geometrically normalized image.  Set to None if the
                               tile is already geometrically transformed.
        @type eye_width:       float or None
        @param tile_size:      The size of the transformed image tile.
        @type  tile_size:      (int, int)
        @param color:          List of 9 floating point values determining the weights of
                               each color [red,green,blue,hue_a,hue_b,saturation,L,a,b].
                               See L{common.colorMix}.
        @param hue_adj:        Adjustment to the hue angle.
        @type  hue_adj:        float
        @param pca_keep:       Number of PCA basis vectors to keep.
        @type  pca_keep:       int
        @param lda_keep:       Number of LDA vectors to keep.
        @type  lda_keep:       int
        @param log_transform:  Log transform teh pixel values.
        @type  log_transform:  True or False
        @param log_shift:      Shift the pixel values by this amound to prevent log(0.0).
        @type  log_shift:      positive float
        @param mean_std:       Use zero mean, one standard deviation pixel normalization.
        @type  mean_std:       True or False
        @param sigma_high:     The high value for band pass filtering.
        @type  sigma_high:     float
        @param sigma_low:      The low value for band pass filtering.
        @type  sigma_low:      float
        @param sim:            Select the type of similarity or distance measure.
        @type sim:             "L1", "L2", or "CORR"
        @param reg:            Regularization parameter for LDA.
        @type  reg:            float
        @param sub_rect:       Extract a sub-rectangle of the transformed image.
                               Mostly for CohortLDA
        @type  sub_rect:       pv.Rect
        @param cohort_adjust:  Normalize the scores using the cohort gallery.
        @type  cohort_adjust:  True or False 
        @param cohort_match:   Number of 'best' cohorts to use for normalization.
        @type  cohort_match:   None or int
        @param feature_norm:   Normalize feature vectors to unit length.
        @type  featuer_norm:   True or False
        @param smoothing:      Smooth the face image before downsampling.  
                               Smoothing is relative to the transformed 
                               image.
        @type  smoothing:      None or float
        
        '''
        # Add a creation time stamp to help identify this training file.
        self.timestamp = time.time()

        # Init parameters
        self.eye_y  = eye_y
        self.eye_width = eye_width
        self.tile_size = tile_size
        self.pca_keep = pca_keep
        self.lda_keep = lda_keep
        self.log_transform = log_transform
        self.mean_std = mean_std
        self.color = color
        self.hue_adj = hue_adj
        self.sigma_low = sigma_low
        self.sigma_high = sigma_high
        self.log_shift = log_shift
        self.sim = sim
        self.reg = reg
        
        self.training_data = []
        self.training_labels = []

        color_mix = np.array(self.color).reshape(9,1,1)
        self.color_mix = pv.unit(color_mix)
        
        self.sub_rect = None
        self.feature_norm = feature_norm
        
        self.cohort_set = []
        self.cohort_adjust = cohort_adjust
        self.cohort_match = cohort_match
        self.smoothing = smoothing
    
    

    def preprocess(self, im, leye, reye):
        '''
        Preprocess the image.  Includes geometric normalization, color channel 
        mixing, band pass filtering, log_transformation, value normalization 
        (meanStd).  B{Note:} eyes should be specified such that leye.X() < reye.X()
        
        @param im: the image to preprocess
        @type im: pv.Image
        @param leye: The left eye coordinate leye.X() < reye.X()
        @type leye: pv.Point
        @param reye: The right eye coordinate
        @type reye: pv.Point
        @returns: the flattend and preprocessed values
        @rtype: np.array
        '''
        
        tile = None
        
        if self.eye_y == None or self.eye_width == None:
            # Assume the image is already geometrically normalized
            assert im.size[0] == self.tile_size[0]
            assert im.size[1] == self.tile_size[1]
            tile = im
        else:
            # geometrically normalize the image
            if leye.X() > reye.X():
                print "Warning, eye coordinates may not be ordered properly."
            w,h = self.tile_size
            left_eye = pv.Point(w*0.5*(1-self.eye_width),self.eye_y*h)
            right_eye = pv.Point(w*0.5*(1+self.eye_width),self.eye_y*h)
            affine = pv.AffineFromPoints(leye, reye, left_eye, right_eye, self.tile_size)
    
            if self.sub_rect != None:
                sw,sh = self.tile_size
                x,y,w,h = self.sub_rect.asTuple()
                rect = pv.Rect(sw*x,sh*y,sw*w,sh*h)
                sub_rect = pv.AffineFromRect(rect,(int(sw*w),int(sh*h)))
                affine = sub_rect*affine
            
            if self.smoothing != None:
                # Compute the size change
                sigma = self.smoothing*leye.l2(reye)/self.eye_width
                
                # Apply smoothing
                im = pv.gaussianFilter(im,sigma)
            
            tile = affine.transformImage(im)
                
        #tile.show(delay=500)
        
        # Separate color channels
        tile = common.colorMix(tile, self.color_mix, self.hue_adj)
        #pv.Image(tile).show(delay=500)
 
        # Band pass filtering   
        if self.sigma_low != None and self.sigma_high != None:     
            sigma_low,sigma_high = max(self.sigma_low,self.sigma_high), min(self.sigma_low,self.sigma_high)
            tile = pv.bandPassFilter(tile, sigma_low, sigma_high)
        
        # log transform
        if self.log_transform:
            signs = np.sign(tile)
            abss = np.abs(tile)
            logs = np.log(abss+self.log_shift)
            tile = signs*logs
        
        # normalize zero mean one std
        if self.mean_std:
            tile = pv.meanStd(tile)
        
        #pv.Image(tile).show(delay=100)
        vec = tile.flatten()
        return vec


    def addTraining(self,label,im,rect,leye,reye,ilog=None):
        '''
        Add a training face.  This is a suggested interface that should 
        work for many algorithms and therefore can use existing training
        code.
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        # Geometric normalization
        vec = self.preprocess(im, leye, reye)
        self.training_data.append(vec)
        self.training_labels.append(label)

    
    def train(self,ilog=None):
        '''
        Train the algorithm.  After all training faces are added this 
        function will be called to do additional computation and processing
        

        @param ilog: optional image log that is used to save intermediate
                        training data such as images of the eigenvectors.
        @type ilog: pv.ImageLog
        '''
        labels = np.array(self.training_labels)
        data = np.array(self.training_data)
        
        # Training PCA
        vals,vecs,mean = pv.pca(data)
        
        # vals should be variance 
        vals = vals**2
        
        if type(self.pca_keep) == int:
            vecs = vecs[:,:self.pca_keep]
        elif type(self.pca_keep) == float and self.pca_keep > 0.0 and self.pca_keep <= 1.0:
            cumenergy = vals.cumsum()
            totenergy = cumenergy[-1]
            idx = cumenergy.searchsorted(self.pca_keep*totenergy)
            #print "keep:",idx,len(cumenergy)
            vecs = vecs[:,:idx]
            self.pca_keep = idx
        else:
            raise NotImplementedError("PCA Keep: %s",self.pca_keep)
        
        self.pca_vecs = vecs
        self.pca_mean = mean
        
        print self.pca_vecs.shape
        
        data = data - mean
        
        data = np.dot(data,vecs)
        
        # Training LDA
        _,vecs,_,_ = pv.lda(data,labels,reg=self.reg)
        
        self.lda_vecs = vecs[:,:self.lda_keep]
        
        self.basis = np.dot(self.pca_vecs,self.lda_vecs)
        
        del self.training_data
        del self.training_labels
        del self.pca_vecs
    
    
    
    def addCohort(self,im,rect,leye,reye,ilog=None):
        '''
        Add a face to use for cohort normalization.  This is a suggested 
        interface that should work for many algorithms.  
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        cohort_record = self.getFaceRecord(im,rect,leye,reye,compute_cohort_scores=False)
        
        self.cohort_set.append(cohort_record)

    
    
    def getFaceRecord(self,im,rect,leye,reye,compute_cohort_scores=True,ilog=None):
        '''
        This function computes a face record.  It takes an image of an 
        unknown face and produces a low dimensional represetation that 
        is used for storage and matching.  The original image can then
        be freed.  For example, in the eigen faces algoirthm this function
        would preprocess the image and then project it onto the eigenbasis.
        What is returned is a face record that contains data used to match
        the face along with other metadata for that image.        
                
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (Not Used)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: an ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        
        @return: A face record object
        @rtype: FaceRecord
        '''
        # Geometric normalization
        vec = self.preprocess(im, leye, reye)
        vec.shape = (1,vec.shape[0])
        
        vec = vec - self.pca_mean
        vec = np.dot(vec,self.basis)
        
        if self.feature_norm:
            vec = pv.unit(vec)
        
        # Create the face record
        fr = common.FaceRecord(rect,leye,reye)
        fr.feature = vec.flatten()

        # Produce cohort statistics
        if compute_cohort_scores:
            scores = self.similarityMatrix([fr], self.cohort_set, cohort_adjust=False)
            scores = scores.flatten()
            
            if self.cohort_match != None:
                order = scores.argsort()
                scores = scores[order]
                scores = scores[-self.cohort_match:]
                
            #print len(scores),self.cohort_match,scores,order
                
            fr.cohort_scores = scores.flatten()
            fr.cohort_mean   = scores.mean()
            fr.cohort_std    = scores.std()
        
        return fr

    
    def similarity(self,face_record1,face_record2):
        '''
        A similarity score computed for two face records.  High values 
        indicate a better match.
        
        @param face_record1: a face record created by calling getFaceRecord.
        @type face_record1: FaceRecord
        
        @param face_record2: a face record created by calling getFaceRecord.
        @type face_record2: FaceRecord
        
        @returns: the similarity as a floating point value
        @rtype: float
        '''
        return self.similarityMatrix([face_record1],[face_record2])[0,0]
    
    
    def similarityMatrix(self,probes,targets,cohort_adjust=None):
        '''
        A similarity matrix computed between sets of face records.  This 
        could be implemented by multiple calls to the "similarity" method, 
        but can be overridden to achieve faster performance when computing
        multiple comparisons.
        
        @param probes: a list of face records.
        @type  probes: [FaceRecord, ...]
        
        @param targets: a list of face records.
        @type  targets: [FaceRecord, ...]
        
        @param cohort_adjust: If not None, then override the cohort adjustment setting
        @type  cohort_adjust: None, True, or False
        
        @return: a similarity matrix in numpy format.
        @rtype: numpy.array
        '''
        if cohort_adjust == None:
            cohort_adjust = self.cohort_adjust
        
        # collect probes and targets in rows
        probe_matrix  = np.array([each.feature for each in probes])
        target_matrix = np.array([each.feature for each in targets])
        
        # compute a similarity matrix
        if self.sim == 'L2':
            l2 = pv.PNorm(2.0)
            scores = -l2(probe_matrix,target_matrix)
        elif self.sim == 'L1':
            l1 = pv.PNorm(1.0)
            scores = -l1(probe_matrix,target_matrix)
        elif self.sim == 'CORR':
            scores = pv.correlation(probe_matrix,target_matrix)
        else:
            raise ValueError("Unknown similarity measure.")
        
        # Perform cohort adjustment
        if cohort_adjust and len(self.cohort_set) > 0: 
            pmeans = np.array([each.cohort_mean for each in probes])
            tmeans = np.array([each.cohort_mean for each in targets])
    
            pstd = np.array([each.cohort_std for each in probes])
            tstd = np.array([each.cohort_std for each in targets])
            
            pmeans.shape = (pmeans.shape[0],1)
            tmeans.shape = (1,tmeans.shape[0])
    
            pstd.shape = (pstd.shape[0],1)
            tstd.shape = (1,tstd.shape[0])
            
            adj = 0.5*(pmeans + tmeans) # Shift
            dnom = 0.5*(pstd + tstd)    # Scale
            
            scores = (scores - adj)/dnom
            
        return scores



class LRLDA:
    '''
    This algorithm is a combination of LDA algorithms run on different local
    regions of the face.  Each region is controlled by the region args.  This
    includes color transformations, normalization, cohort norm, etc. The
    initial normalized rectangle is governed by the parameters here.
    '''
    def __init__(self,region_args, 
                    eye_y = 0.35,
                    eye_width = 0.55,
                    tile_size = (65,85),
                    smoothing=None,
                    ):
        '''
        Create an CohortLDA algorithm.
        
        @param region_args: a list of regions and there parameters for LDA.
        @type  region_args: list
        @param eye_y:          The y location of the eyes relative to the height of 
                               the geometrically normalized image.  Set to None if the
                               tile is already geometrically transformed.
        @type eye_y:           float or None
        @param eye_width:      The width location of the eyes relative to the width of 
                               the geometrically normalized image.  Set to None if the
                               tile is already geometrically transformed.
        @type eye_width:       float or None
        @param tile_size:      The size of the transformed image tile.
        @type  tile_size:      (int, int)
        @param smoothing:      Smooth the face image before downsampling.  
                               Smoothing is relative to the transformed 
                               image.
        @type  smoothing:      None or float
        '''
        
        # Add a creation time stamp to help identify this training file.
        self.timestamp = time.time()

        self.eye_y = eye_y
        self.eye_width = eye_width
        self.tile_size = tile_size
        
        self.smoothing = smoothing
        
        self.regions = []
        for kwargs in region_args:
            # make sure regios are not used 
            kwargs['eye_y']     = None
            kwargs['eye_width'] = None
            rect = kwargs['rect']
            tile_size = kwargs['tile_size']
            del kwargs['rect']
            
            alg = LDA(**kwargs)
            
            self.regions.append( (rect,tile_size,alg) )
            
            
            
    def preprocess(self, im, leye, reye):
        '''
        Preprocess the image.  Includes geometric normalization, color channel 
        mixing, band pass filtering, log_transformation, value normalization 
        (meanStd).  B{Note:} eyes should be specified such that leye.X() < reye.X()
        
        @param im: the image to preprocess
        @type im: pv.Image
        @param leye: The left eye coordinate leye.X() < reye.X()
        @type leye: pv.Point
        @param reye: The right eye coordinate
        @type reye: pv.Point
        @returns: a geometrically normalized image
        @rtype: pv.Image
        '''
        
        # geometrically normalize the image
        if leye.X() > reye.X():
            print "Warning, eye coordinates may not be ordered properly."
        w,h = self.tile_size
        left_eye = pv.Point(w*0.5*(1-self.eye_width),self.eye_y*h)
        right_eye = pv.Point(w*0.5*(1+self.eye_width),self.eye_y*h)
        affine = pv.AffineFromPoints(leye, reye, left_eye, right_eye, self.tile_size)
        
        if self.smoothing != None:
            # Compute the size change
            sigma = self.smoothing*leye.l2(reye)/(self.eye_width*w)
            
            # Apply smoothing
            im = pv.gaussianFilter(im,sigma)

        tiles = []
        for rect,tile_size,_ in self.regions:
            x = rect.x
            y = rect.y
            w = rect.w
            h = rect.h
            tw,th = self.tile_size
            rect = pv.Rect(x*tw,y*th,w*tw,h*th)
            tile = (pv.AffineFromRect(rect,tile_size)*affine).transformImage(im)
            tiles.append(tile)
            
        return tiles
    
    
    
    def addTraining(self,label,im,rect,leye,reye,ilog=None):
        '''
        Add a training face.  This is a suggested interface that should 
        work for many algorithms and therefore can use existing training
        code.
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        # Geometric normalization
        
        tiles = self.preprocess(im, leye, reye)
        for i in range(len(self.regions)):
            _,_,alg = self.regions[i]
            tile = tiles[i]
            alg.addTraining(label,tile,None,None,None)

    
    def train(self,ilog=None):
        '''
        Train the algorithm.  After all training faces are added this 
        function will be called to do additional computation and processing
        

        @param ilog: optional image log that is used to save intermediate
                        training data such as images of the eigenvectors.
        @type ilog: pv.ImageLog
        '''
        for i in range(len(self.regions)):
            _,_,alg = self.regions[i]
            alg.train()
    
    
    
    def addCohort(self,im,rect,leye,reye,ilog=None):
        '''
        Add a face to use for cohort normalization.  This is a suggested 
        interface that should work for many algorithms.  
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        tiles = self.preprocess(im, leye, reye)
        for i in range(len(self.regions)):
            _,_,alg = self.regions[i]
            tile = tiles[i]
            alg.addCohort(tile,None,None,None)

    
    
    def getFaceRecord(self,im,rect,leye,reye,ilog=None):
        '''
        This function computes a face record.  It takes an image of an 
        unknown face and produces a low dimensional represetation that 
        is used for storage and matching.  The original image can then
        be freed.  For example, in the eigen faces algoirthm this function
        would preprocess the image and then project it onto the eigenbasis.
        What is returned is a face record that contains data used to match
        the face along with other metadata for that image.        
                
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (Not Used)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: an ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        
        @return: A face record object
        @rtype: FaceRecord
        '''
        
        rec = common.FaceRecord(rect,leye,reye)
        rec.features = []
        
        # Geometric normalization
        tiles = self.preprocess(im, leye, reye)
        for i in range(len(self.regions)):
            rect,_,alg = self.regions[i]
            tile = tiles[i]
            subrec = alg.getFaceRecord(tile,None,None,None)
            rec.features.append(subrec)
        
        
        return rec

    
    def similarity(self,face_record1,face_record2):
        '''
        A similarity score computed for two face records.  High values 
        indicate a better match.
        
        @param face_record1: a face record created by calling getFaceRecord.
        @type face_record1: FaceRecord
        
        @param face_record2: a face record created by calling getFaceRecord.
        @type face_record2: FaceRecord
        
        @returns: the similarity as a floating point value
        @rtype: float
        '''
        return self.similarityMatrix([face_record1],[face_record2])[0,0]
    
    
    def similarityMatrix(self,probes,targets,cohort_adjust=None):
        '''
        A similarity matrix computed between sets of face records.  This 
        could be implemented by multiple calls to the "similarity" method, 
        but can be overridden to achieve faster performance when computing
        multiple comparisons.
        
        @param probes: a list of face records.
        @type  probes: [FaceRecord, ...]
        
        @param targets: a list of face records.
        @type  targets: [FaceRecord, ...]
        
        @param cohort_adjust: If not None, then override the cohort adjustment setting
        @type  cohort_adjust: None, True, or False
        
        @return: a similarity matrix in numpy format.
        @rtype: numpy.array
        '''
        
        mat = np.zeros((len(probes),len(targets)),dtype=np.float32)
        # Geometric normalization
        for i in range(len(self.regions)):
            _,_,alg = self.regions[i]
            probe_tmp  = [each.features[i] for each in probes]
            target_tmp = [each.features[i] for each in targets]
            tmp = alg.similarityMatrix(probe_tmp,target_tmp)
            mat += tmp
            
        mat = mat / len(self.regions)    
            
        return mat
    
    
    
    
    
    
    
    
