# Copyright (c) 2010 Colorado State University and David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Contains classes and supporting data for the Local Region Principal 
Components Analysis (LRPCA) algorithm.

@author: bolme
'''
import  pyvision as pv
from pca import PCA, FaceRecord
import numpy as np
import time



DEFAULT_REGIONS = [
    [pv.Point(0.5,0.5),1.0],
    [pv.Point(0.25,0.33),0.33],
    [pv.Point(0.75,0.33),0.33],
    [pv.Point(0.18,0.18),0.33],
    [pv.Point(0.32,0.18),0.33],
    [pv.Point(0.68,0.18),0.33],
    [pv.Point(0.82,0.18),0.33],
    [pv.Point(0.50,0.33),0.33],
    [pv.Point(0.50,0.50),0.33],
    [pv.Point(0.40,0.65),0.33],
    [pv.Point(0.60,0.65),0.33],
    [pv.Point(0.35,0.90),0.20],
    [pv.Point(0.50,0.90),0.20],
    [pv.Point(0.65,0.90),0.20],
                   ]
'''
This list defines the regions used for the default version of LRPCA.

The list contains the center point of the region and the width of the
square used for the feature.  All coordinates are relative to tile defined 
by the eyes and tile size coordinate.  For example, pv.Point(0.5, 0.5) is
the center of the face rectangle and 1.0 defines the width to be the total 
width of the face tile.   

This particular list includes 14 regions where the first region is the entire 
face.
'''

DEFAULT_REGIONS2 = [
    #[pv.Point(0.5,0.5),1.0],
    [pv.Point(0.25,0.33),0.33],
    [pv.Point(0.75,0.33),0.33],
    [pv.Point(0.18,0.18),0.33],
    [pv.Point(0.32,0.18),0.33],
    [pv.Point(0.68,0.18),0.33],
    [pv.Point(0.82,0.18),0.33],
    [pv.Point(0.50,0.33),0.33],
    [pv.Point(0.50,0.50),0.33],
    [pv.Point(0.40,0.65),0.33],
    [pv.Point(0.60,0.65),0.33],
    [pv.Point(0.35,0.90),0.20],
    [pv.Point(0.50,0.90),0.20],
    [pv.Point(0.65,0.90),0.20],
                   ]
'''Regions without full face'''

RIGHT_OCULAR_REGIONS = [
    #[pv.Point(0.5,0.5),1.0],
    [pv.Point(0.25,0.33),0.33],
    #[pv.Point(0.75,0.33),0.33],
    [pv.Point(0.18,0.18),0.33],
    [pv.Point(0.32,0.18),0.33],
    #[pv.Point(0.68,0.18),0.33],
    #[pv.Point(0.82,0.18),0.33],
    #[pv.Point(0.50,0.33),0.33],
    #[pv.Point(0.50,0.50),0.33],
    #[pv.Point(0.40,0.65),0.33],
    #[pv.Point(0.60,0.65),0.33],
    #[pv.Point(0.35,0.90),0.20],
    #[pv.Point(0.50,0.90),0.20],
    #[pv.Point(0.65,0.90),0.20],
                   ]

LEFT_OCULAR_REGIONS = [
    #[pv.Point(0.5,0.5),1.0],
    #[pv.Point(0.25,0.33),0.33],
    [pv.Point(0.75,0.33),0.33],
    #[pv.Point(0.18,0.18),0.33],
    #[pv.Point(0.32,0.18),0.33],
    [pv.Point(0.68,0.18),0.33],
    [pv.Point(0.82,0.18),0.33],
    #[pv.Point(0.50,0.33),0.33],
    #[pv.Point(0.50,0.50),0.33],
    #[pv.Point(0.40,0.65),0.33],
    #[pv.Point(0.60,0.65),0.33],
    #[pv.Point(0.35,0.90),0.20],
    #[pv.Point(0.50,0.90),0.20],
    #[pv.Point(0.65,0.90),0.20],
                   ]

GBU_TUNING = {
              'left_eye': 128*pv.Point(1./4.,1./3.),
              'right_eye': 128*pv.Point(3./4.,1./3.),
              'tile_size': (128,128),
              'regions': DEFAULT_REGIONS,
              'fisher_thresh': 3500, # keep all 3500 basis vectors
              'self_quotient': 3.0, 
              # Training GBU Uncontrolled x2
              }
'''
This dictionary is a good tuning for the Good, Bad, and Ugly dataset.  

This tuning can be applied to the Local Region PCA when it is created
using a call like this:

C{> face_rec = LRPCA(**GBU_TUNING)}

I{Note: This is intended to be used with the GBU_Training_Uncontrolledx2.xml training sigset and reduced images.}

Ugly: 7.3%
Good: 24%
Ugly: 64%
'''

LEFT_OCULAR_TUNING = {
              'left_eye': 128*pv.Point(1./4.,1./3.),
              'right_eye': 128*pv.Point(3./4.,1./3.),
              'tile_size': (128,128),
              'regions': LEFT_OCULAR_REGIONS,
              'fisher_thresh': 750, # keep all 750 basis vectors
              'self_quotient': 3.0, 
              # Training GBU Uncontrolled x2
              }
'''
LRPCA setup for left ocular recognition.
'''

RIGHT_OCULAR_TUNING = {
              'left_eye': 128*pv.Point(1./4.,1./3.),
              'right_eye': 128*pv.Point(3./4.,1./3.),
              'tile_size': (128,128),
              'regions': RIGHT_OCULAR_REGIONS,
              'fisher_thresh': 750, # keep all 750 basis vectors
              'self_quotient': 3.0, 
              # Training GBU Uncontrolled x2
              }
'''
LRPCA setup for right ocular recognition.
'''



class LRPCA:
    '''
    This class implements a Local Region Principal Components Analysis 
    algorithm.
    
    Training and testing images are first preprocessed.   First a similarity 
    transformation is applied which rotates, scales, and translates the faces
    to align the eyes.  The image is then cropped tightly around the face to 
    for a face tile. 
    
    The algorithm then splits the face into a small number of regions
    and then runs PCA on each.  This provides basis vectors for each facial
    region.  During training a weight is also learned for each eigenvector in 
    each region.  This weight determines the influence of each vector during 
    testing and is based on the Fisher Criterion.
    
    When testing, each face region is projected onto the basis vectors which 
    generates a set of coefficients.  GBU Training, for example, has 14 regions 
    with 250 vectors each.  This results is 3500 coefficients.  Each coefficient
    is weighted using the Fisher Criterion.  The similarity of two vectors is
    computed using Pearson's normalized correlation.
    
    This algorithm does NOT adjust the regions to recenter them on facial
    features.  It does reweight the regions and coefficients independently.
    '''
    
    def __init__(self,
                 left_eye  = 128*pv.Point(1./4.,1./3.),
                 right_eye = 128*pv.Point(3./4.,1./3.),
                 tile_size = (128,128),
                 regions = DEFAULT_REGIONS,
                 fisher_thresh = .50,
                 self_quotient = 3.0, 
                 **kwargs
                 ):
        '''
        @param left_eye: the standard location for the left eye coordinate in the image tile.
        @type left_eye: pv.Point
        
        @param right_eye: the standard location for the right eye coordinate in the image tile.
        @type right_eye: pv.Point
        
        @param tile_size: the size of the geometrically normalized tile.
        @type tile_size: [width,height]
        
        @param regions: a list of the local regions defined by the center point and width.  
        @type regions: L{DEFAULT_REGIONS}
        
        @param fisher_thresh: determines which vectors to keep.  If this is a floating point 
            value it is a threshold for the Fisher Criterion.  If it is an int it is the total 
            number of vectors.
        @type fisher_thresh: int|float
        
        @param self_quotient: the radius of the gaussian used for self quotient image.
        @type self_quotient: float
        
        @param kwargs: Additional keyword arguments are passed on to the PCA algorithms.
        '''
        # Add a creation time stamp to help identify this training file.
        self.timestamp = time.time()
        
        # These define the transformation used for the face tile.
        self.left_eye = left_eye
        self.right_eye = right_eye
        self.tile_size = tile_size
        
        # This defines the geometry of the face regions 
        self.regions = regions
        
        # This defines the minimum fisher threshold to keep the eigenvector
        self.fisher_thresh = fisher_thresh
        
        # This defines the radius of the Gaussian used in the Self Quotient Image normalization
        self.self_quotient = self_quotient

        # These are the weights for the Fisher Criterion
        self.weight = None
        
        # Each region is handled by its own pca algorithm.  This could
        # probably be done better given more time but for now this just
        # works.
        self.pcas = []
        for region in self.regions:
            rect = self._regionToRect(region)
            ful = pv.Point(0,0)
            flr = pv.Point(rect.w,rect.h)
            ts = (rect.w,rect.h)
            
            self.pcas.append(PCA(left_eye=ful, right_eye=flr, tile_size=ts, self_quotient=self_quotient, **kwargs))
            
        self.training_tiles = []
            
        
    def _regionToRect(self,region):
        '''
        Convert a region description (center,width) into a rectangle. 
        This should not normally be called from outside this class.
        
        @rtype: pv.Rect
        '''
        w = self.tile_size[0]
        rect = pv.Rect(w*region[0].X()-0.5*w*region[1],w*region[0].Y()-0.5*w*region[1],w*region[1],w*region[1])
        return rect 
        
    def preprocess(self,im,rect,leye,reye,ilog=None):
        '''
        For LRPCA this just extracts the image tile.  If an ilog
        is given, an image will be saved showing the local
        regions.  This should not normally be called from outside this class.
        
        @param im: an input image.
        @type im: pv.Image

        @param rect: the face detection rectangle (not used)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate
        @type leye: pv.Point
        
        @param reye: the right eye coordinate
        @type reye: pv.Point
        
        @param ilog: an optional PyVision ImageLog
        @type ilog: pv.ImageLog
        
        @return: the geometrically normalized tile
        @rtype: pv.Image
        ''' 
        # Geometric normalization
        affine = pv.AffineFromPoints(leye,reye,self.left_eye,self.right_eye,self.tile_size)
        
        tile = affine.transformImage(im)
        
        pil = tile.asPIL()
        pil = pil.convert('L')
        
        tile = pv.Image(pil)
        if ilog != None:
            ilog.log(tile,label="EigenfacesOrig",format='jpg')
            for region in self.regions:
                rect = self._regionToRect(region)
                tile.annotateEllipse(rect)
            ilog.log(tile,label="EigenfacesPreprocessedTraining",format='jpg')
                
        # Return the final vector
        return tile

    
    def addTraining(self,label,im,rect,leye,reye,ilog=None):
        '''
        When training data is added to this algorithm it breaks the 
        image into the local regions and then adds those image tiles
        to the training data for individual PCA algorithms.

        Additional preprocessing is performed by each PCA algorithm....
        First the image is geometrically normalized. The eye coordinates
        are used to rotate, scale, and translate the images so that the
        eyes are at standard positions.  The image is also cropped to the 
        specified size.
        
        Next, the image is vectorized. Such that each value in the image 
        is corresponds to value in the vector.  
        
        Finally the image is value normalized.  The pixel values are rescaled to 
        have a mean of 0.0 and a standard deviation of 1.0.
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        tile = self.preprocess(im,rect,leye,reye,ilog=ilog)
        
        self.training_tiles.append([label,tile])
        
        for r in range(len(self.regions)):
            region = self.regions[r]
            pca = self.pcas[r]
            rect = self._regionToRect(region)
            
            sul = pv.Point(rect.x,rect.y)
            slr = pv.Point(rect.x+rect.w,rect.y+rect.h)
            
            pca.addTraining(label, tile, None, sul, slr, ilog=ilog)
            
    def train(self,ilog=None):
        '''
        Call this method after all the training data has been added.
        This calls the training function for each of the individual PCA
        algorithms.  Each PCA can then be used to project the local regions
        to lower dimensional representations with data whitening. 
        
        Once the individual algorithms are trained each image in the training
        set is projected into the new eigenspaces.  The coeffecents from all 
        region projections are then concatentated into one long vector per
        face.  The  coeffecients in those vectors are then reweighted using 
        the Fisher Criterion computed on the training set.
        
        @param ilog: optional image log that is used to save intermediate
                        training data such as images of the eigenvectors.
        @type ilog: pv.ImageLog
        '''
        
        # Train each pca method individually
        for r in range(len(self.regions)):
            #print "Training PCA",r
            pca = self.pcas[r]
            pca.train(ilog=ilog)
            
        # Project each training face using the getFaceRecord method.
        training_recs = []
        for label,tile in self.training_tiles:
            fr = self.getFaceRecord(tile,None,self.left_eye,self.right_eye)
            training_recs.append([label,fr.feature])
            n = len(fr.feature)
                    
        # compute the class means
        class_n = {}
        class_means = {}
        for lab,vec in training_recs:
            if not class_n.has_key(lab):
                class_n[lab] = 0
                class_means[lab] = np.zeros(vec.shape,dtype=np.float64)
                
            class_n[lab] += 1
            class_means[lab] += vec
            
        # compute the between class variance for each coefficient
        sbm = []
        for lab in class_n.keys():
            class_means[lab] /= class_n[lab]
            sbm.append(class_means[lab])
                
        sbm = np.array(sbm)
        sb = sbm.var(axis=0)
        
        # compute the within class variance for each coefficient
        swm = []
        for lab,vec in training_recs:
            swm.append(vec-class_means[lab])
        
        swm = np.array(swm)
        
        sw = swm.var(axis=0)
        
        # compute the weight for each coefficient using the Fisher Criterion (between var/within var)
        self.weight = sb/sw
        
        # Keep only N basis vectors.
        if isinstance(self.fisher_thresh,int):
            tmp = self.weight.copy()
            tmp.sort()
            if self.fisher_thresh > len(tmp):
                self.fisher_thresh = len(tmp)
            self.fisher_thresh = tmp[-self.fisher_thresh]
        
        keep = (self.weight >= self.fisher_thresh)
        #print "Weight:",self.weight, ( self.weight >= self.fisher_thresh ).sum()
        
        # Drop coefficients and basis vectors that were not kept.
        self.weight = self.weight[keep]

        tmp = keep.copy()
        for pca in self.pcas:
            n,m = pca.eigenbasis.shape
        
            pca.eigenbasis = pca.eigenbasis[tmp[:n],:]
            pca.whitenvalues = pca.whitenvalues[tmp[:n]]
            tmp = tmp[n:]
                    
        # Clean up unneeded training data
        self.training_tiles = []
        
        

            
            
            
    def getFaceRecord(self,im,rect,leye,reye,ilog=None):
        '''
        During testing a face record is computed for each  face.  This record is typically
        a few hundred values that are representitive of the images 
        appearance.  The first step is to create a smaller image region
        for each facial feature.  Once that is created each region is preprocessed 
        to produce a normalized vector.  The region vectors are then 
        projected onto the eigenbasis.  If whitening 
        is enabled the components are rescaled to have a uniform 
        distribution.  The face record is then created by concatenating each of 
        the projected vectors.  The values in this record are then reweighed 
        using the Fisher Criterion.
        
                
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (Not Used)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: an ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        
        @return: A face record object
        @rtype: FaceRecord
        '''
        # preprocess the image
        tile = self.preprocess(im,rect,leye,reye,ilog=ilog)
        
        vecs = []

        for r in range(len(self.regions)):
            region = self.regions[r]
            pca = self.pcas[r]
            rect = self._regionToRect(region)
            
            sul = pv.Point(rect.x,rect.y)
            slr = pv.Point(rect.x+rect.w,rect.y+rect.h)
            
            fr = pca.getFaceRecord(tile, None, sul, slr, ilog=None)
            
            vecs.append(fr.feature)
            
        # Concatinate everything into a single vector
        vec = np.concatenate(vecs)
        
        # Reweight the vectors using the Fisher Criterion
        if self.weight != None:
            vec = self.weight*vec
        
        # return the face record
        fr = FaceRecord(rect,leye,reye)
        fr.feature = vec
        
        return fr
        
    
    def similarity(self,face_record1,face_record2):
        '''
        A similarity score is computed using Pearson's correlation between
        the face records.
        
        @param face_record1: a face record created by calling getFaceRecord.
        @type face_record1: FaceRecord
        
        @param face_record2: a face record created by calling getFaceRecord.
        @type face_record2: FaceRecord
        
        @returns: the similarity as a floating point value
        @rtype: float
        '''
        return self.similarityMatrix([face_record1],[face_record2])[0,0]

        
    def similarityMatrix(self,probes,targets):
        '''
        A similarity matrix is computed using Pearson's correlation between
        the face records in the probes and targets lists.
        
        @param probes: a list of face records.
        @type probes: [FaceRecord, ...]
        
        @param targets: a list of face records.
        @type targets: [FaceRecord, ...]
        
        
        @return: a similarity matrix in numpy format.
        @rtype: numpy.array
        '''
        # collect probes and targets in rows
        probe_matrix = []
        for each in probes:
            probe_matrix.append(each.feature)
        probe_matrix  = np.array(probe_matrix)
        
        target_matrix = []
        for each in targets:
            target_matrix.append(each.feature)
        target_matrix  = np.array(target_matrix)

        # compute a correlation matrix
        tmp = pv.correlation(probe_matrix,target_matrix)
        
        return tmp
    
    def __str__(self):
        '''
        A string containing basic info about this algorithm.
        '''
        return "LRPCA(Regions:%d, Dimensions:%d)"%(len(self.regions),len(self.weight))

