    # Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Contains classes and supporting data for a Principal Components
Analysis (PCA) algorithm.

This is a reference implementation of an Eigenfaces type face 
recognition algorithm.

This algorithm is based on the techniques described in the following 
papers.

M. A. Turk and A. P. Pentland. Face Recognition Using Eigenfaces. 
Computer Vision and Pattern Recognition. 1991.

M. Kirby and L. Sirovich, Application of the Karhunen-Loeve Procedure 
for the Characterization of Human Faces. Trans. on Pattern Analysis 
and Machine Intelligence. vol. 12, pp. 103-107. January 1990.

This implementation has a number of modifications that improve the 
accuracy of the Eigenfaces algorithm.   
'''

import pyvision as pv
import numpy as np
import scipy as sp
import scipy.linalg
import time

class FaceRecord:
    '''This is a simple structure to organize information on a face.'''
    
    def __init__(self, face_rect, left_eye, right_eye):
        self.face_rect = face_rect
        self.left_eye  = left_eye
        self.right_eye = right_eye

PCA_FERET_TUNING = {
                 'left_eye'      : pv.Point(32.0,64.0),
                 'right_eye'     : pv.Point(96.0,64.0),
                 'tile_size'     : (128,128),
                 'mask'          : None,
                 'value_norm'    : pv.meanStd,
                 'self_quotient' : 11.0,
                 'whiten'        : True,
                 'drop'          : 2,
                 'keep'          : 250,
                 }
'''
Tuning that works well for the FERET dataset.
'''

PCA_FRGC_TUNING = {
                 'left_eye'      : pv.Point(32.0,64.0),
                 'right_eye'     : pv.Point(96.0,64.0),
                 'tile_size'     : (128,128),
                 'mask'          : None,
                 'value_norm'    : pv.meanStd,
                 'self_quotient' : 11.0,
                 'whiten'        : True,
                 'drop'          : 2,
                 'keep'          : 0.95,
                 }
'''
Tuning that works well for FRGC.
'''


PCA_GBU_TUNING = {
                 'left_eye'      : pv.Point(32.0,64.0),
                 'right_eye'     : pv.Point(96.0,64.0),
                 'tile_size'     : (128,128),
                 'mask'          : None,
                 'value_norm'    : pv.meanStd,
                 'self_quotient' : 11.0,
                 'whiten'        : True,
                 'drop'          : 2,
                 'keep'          : 250,
                 }
'''
Tuning that works well for GBU.
'''



class PCA:
    
    def __init__(self,
                 left_eye  = pv.Point(32.0,64.0),
                 right_eye = pv.Point(96.0,64.0),
                 tile_size = (128,128),
                 mask = None,
                 value_norm = pv.meanStd,
                 self_quotient = 11.0,
                 whiten = True,
                 drop = 2,
                 keep = 250,
                 ):
        '''
        Initialize the eigenfaces class.
        
        @param self_quotient: The gausian radius (sigma) for the self quotient image. Set to None to disable.
        @type self_quotient: float or None
        '''
        # Add a creation time stamp to help identify this training file.
        self.timestamp = time.time()

        # Geometric parameters
        self.left_eye  = left_eye
        self.right_eye = right_eye
        self.tile_size = tile_size
        
        if mask != None:
            raise NotImplementedError("Masks are not supported yet in this version of PCA")
        
        self.mask = mask
        
        
        # Normalization parameters
        self.value_norm = value_norm
        
        # Sigma for the self quotient
        self.self_quotient = self_quotient
        
        # Eigenbases parameters
        self.whiten = whiten
        self.drop   = drop
        self.keep   = keep
        
        # Setup training vectors
        self.training_vectors = []
        
        # Eigenbasis Information
        self.mean = None
        
    def addTraining(self,label,im,rect,leye,reye,ilog=None):
        '''
        Add a training face.
        
        @param label: a subject identifier for this face.
        @type label: int | str
        
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (unused)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: optional ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        '''
        vec = self.preprocess(im,rect,leye,reye,ilog=ilog)

        self.training_vectors.append([label,vec])
    
    def train(self,ilog=None):
        '''
        Train the eigenfaces algorithm.

        @param ilog: optional image log that is used to save intermediate
                        training data such as images of the eigenvectors.
        @type ilog: pv.ImageLog
        '''
        assert len(self.training_vectors) > 1
                
        # Build data matrix
        train_vecs = [vec for lab,vec in self.training_vectors]
        X = np.array(train_vecs).transpose()
        n,m = X.shape
        self.mean = X.mean(axis=1)
        n = self.mean.shape[0]
        self.mean = self.mean.reshape(n,1)
        
        if ilog != None:
            ilog.log(pv.Image(self.mean.reshape(self.tile_size)),label="EigenfacesMeanFace")

        # Mean subtract the data vectors
        X = X - self.mean
        
        # Make sure the svd is done in 64 bits
        X = np.array(X,dtype=np.float64) 
        
        # Compute PCA using the SVD method
        U,D,Vt = sp.linalg.svd(X,full_matrices=0)
        
        # Normalize the covariance
        D = D/np.sqrt(m)
        w = D**2 # compute the eigen values
        
        # compute the total energy
        self.total_energy = w.sum()
        
        # drop large eigenvalues and vectors from the front
        if self.drop > 0:
            w = w[self.drop:]
            D = D[self.drop:]
            U = U[:,self.drop:]
        
        # compute the remaining energy
        energy = w.sum()
        cumenergy = w.cumsum()
        
        # drop small energies
        if type(self.keep) == float:
            keep = self.keep*energy
            cutoff = cumenergy.searchsorted(keep)
        elif type(self.keep) == int:
            cutoff = self.keep
        else:
            raise TypeError("Unexpected argument type for keep: %s"%type(self.keep))
        cumenergy = cumenergy[:cutoff]
        w = w[:cutoff]
        D = D[:cutoff]
        U = U[:,:cutoff]
        self.final_energy = w.sum()
        
        # compute the eigenbasis
        self.eigenbasis   = U.transpose()
        self.eigenvalues  = w
        self.whitenvalues = 1.0/D

        # Clear training vectors
        self.training_vectors = []

        # Perform self checks on the pca process
        n,m = U.shape
        if np.abs(np.dot(U.transpose(),U)-np.eye(m)).max() > 0.0001:
            raise ValueError("EigenBasis is not orthonormal: %f"%np.abs(np.dot(U.transpose(),U)-np.eye(m)).max())
        
        if ilog != None:
            eigenfaces = self.getEigenfaces()
            for im in eigenfaces:
                ilog.log(im,label="EigenfaceBasis")

        
    def preprocess(self,im,rect,leye,reye,ilog=None):
        '''
        This function preprocesses the face tile.  
        
        First the image is geometrically normalized. The eye coordinates
        are used to rotate, scale, and translate the images so that the
        eyes are at standard positions.  The image is also cropped to the 
        specified size.
        
        The self quotient image method of Wang et.al \cite{Wang} is applied
        using a Gaussian filter of radius 3.0.  This has the effect of reducing
        the effect of lighting.

        Next, the image is vectorized. Such that each value in the image 
        is corresponds to value in the vector.  
        
        
        Finally the image is value normalized.  The pixel values are rescaled to 
        have a mean of 0.0 and a standard deviation of 1.0.
        
        @param im: an input image.
        @type im: pv.Image

        @param rect: the face detection rectangle (not used)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate
        @type leye: pv.Point
        
        @param reye: the right eye coordinate
        @type reye: pv.Point
        
        @param ilog: an optional PyVision ImageLog
        @type ilog: pv.ImageLog
        
        @return: the normalized image
        @rtype: numpy.vector
        ''' 
        # Geometric normalization
        affine = pv.AffineFromPoints(leye,reye,self.left_eye,self.right_eye,self.tile_size)
        tile = affine.transformImage(im)
        
        if self.self_quotient != None:
            tile = pv.selfQuotientImage(tile,sigma = self.self_quotient)
                    
        if ilog != None:
            ilog.log(tile,label="EigenfacesPreprocessedTraining")
                
        # Vectorize
        vec = tile.asMatrix2D().flatten()
        
        # Value Normalization
        vec = self.value_norm(vec)
        
        # Return the final vector
        return vec
    
    
    def getFaceRecord(self,im,rect,leye,reye,ilog=None):
        '''
        This function computes a face record.  This record is typically
        a few hundred values that are representitive of the images 
        appearance.  The image is first preprocessed to produce a normalized 
        vector.  The vector is projected onto the eigenbasis.  If whitening 
        is selected the components are rescaled to have a uniform 
        distribution.  The face record consists of this projected vector 
        and some additional image metadata.
        
                
        @param im: an image containing a face.
        @type im: pv.Image
        
        @param rect: a face detection rectangle. (Not Used)
        @type rect: pv.Rect
        
        @param leye: the left eye coordinate.
        @type leye: pv.Point
        
        @param reye: the right eye coordinate.
        @type reye: pv.Point
        
        @param ilog: an ImageLog used for saving intermediate data.
        @type ilog: pv.ImageLog
        
        @return: A face record object
        @rtype: FaceRecord
        '''
        n,m = self.eigenbasis.shape
        
        # preprocess the image
        vec = self.preprocess(im,rect,leye,reye,ilog=ilog)
        vec = vec.reshape(m,1)
        
        # subtract the mean
        vec = vec - self.mean
        
        # project onto the basis
        vec = np.dot(self.eigenbasis,vec)
        vec = vec.flatten()
        
        # whiten if necessary
        if self.whiten:
            vec = self.whitenvalues*vec
        
        # return the face record
        fr = FaceRecord(rect,leye,reye)
        fr.feature = vec
        
        return fr
        
    
    def similarity(self,face_record1,face_record2):
        '''
        A similarity score is computed using Pearson's correlation between
        the face records.
        
        @param face_record1: a face record created by calling getFaceRecord.
        @type face_record1: FaceRecord
        
        @param face_record2: a face record created by calling getFaceRecord.
        @type face_record2: FaceRecord
        
        @returns: the similarity as a floating point value
        @rtype: float
        '''
        return self.similarityMatrix([face_record1],[face_record2])[0,0]
        
        
    def similarityMatrix(self,probes,targets):
        '''
        A similarity matrix is computed using Pearson's correlation between
        the face records in the probes and targets lists.
        
        @param probes: a list of face records.
        @type probes: [FaceRecord, ...]
        
        @param targets: a list of face records.
        @type targets: [FaceRecord, ...]
        
        @return: a similarity matrix in numpy format.
        @rtype: numpy.array
         '''
        # collect probes and targets in rows
        probe_matrix  = np.array([each.feature for each in probes])
        target_matrix = np.array([each.feature for each in targets])
        
        # compute a correlation matrix
        tmp = pv.correlation(probe_matrix,target_matrix)
        
        return tmp


    def backProject(self,face_record):
        '''
        Use the eigenbasis to back project a face record and obtain a reconstructed image.
        
        @param face_record: A face record object
        @type face_record: FaceRecord
        
        @return: the back projected face image
        @rtype: numpy.array
        '''
        vec = face_record.feature
        
        # dewhiten if necessary
        if self.whiten:
            vec = vec/self.whitenvalues
        r = vec.shape[0]
        vec = vec.reshape(r,1)

        # project onto the basis
        vec = np.dot(self.eigenbasis.transpose(),vec)

        # subtract the mean
        vec = vec + self.mean
        
        vec = vec.reshape(self.tile_size)

        return pv.Image(vec)
    
    
    def getEigenfaces(self):
        '''
        Returns the eigenvectors as a list of images.
        
        @return: EigenFaces
        @rtype: [pv.Image, ...]

        '''
        eigenfaces = []
        for each in self.eigenbasis:
            tile = each.reshape(self.tile_size)
            eigenfaces.append(pv.Image(tile))
            
        return eigenfaces
