# Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
Created on Sep 28, 2011

@author: bolme
'''
import pyvision as pv
import scipy.ndimage as ndi
import numpy as np

def prcomp(data,center=False,scale=False):
    '''
    observations are in columns.
    
    This function should yeild results similar to R.
    
    @returns: stdevs,rotation
    '''
    data = data.copy()
    r,c= data.shape
    
    if center:
        ctr = data.mean(axis=0).reshape(1,c)
        data = data - ctr 
    
    if scale:
        scl = data.std(axis=0).reshape(1,c)
        data = data/scl 
        
    cov = np.dot(data.transpose(),data)
    u,val,vt = np.linalg.svd(cov)
    #print cov
    #print val
    standard_dev = np.sqrt(val/r)

    result = [standard_dev,u]
    
    if center:
        result.append(ctr)
        
    if scale:
        result.append(scl)
        
    return result
    
    
def focusMaxGradient(im, der=1, normalize=True, quantile=0.975, full_output=False):
    '''
    This is a measure estimates the focus of an image.  High values 
    sharper images.  The measure seems to be fairly robust to changes in 
    illumination and contrast.
    
    Set grid to None no find the maximum curvature for the entire tile.
    Grid will subdivide the image into a grid and take the median of the 
    maximum of each subimage which may be more robust than the global 
    maximum. 
    
    For the standard measure the face should be geometrically normalized 
    to have eyes at (60,80) and (196,80) with dimensions 256x256.
    
    affine = pv.AffineFromPoints(eye1, eye2, pv.Point(60.0,80.0), pv.Point(196.0,80.0), (256,256))

    
    Algorithm by David Bolme and Yui Man Lui
    
    @param im: an image.
    '''
    mat = im.asMatrix2D()
    
    if normalize:
        mat = pv.meanStd(mat)
    
    # Sobel twice for second derivative
    if der == 2:
        sdx = ndi.sobel(ndi.sobel(mat,axis=0),axis=0)
        sdy = ndi.sobel(ndi.sobel(mat,axis=1),axis=1)
    elif der == 1:
        sdx = ndi.sobel(mat,axis=0)
        sdy = ndi.sobel(mat,axis=1)
    else:
        raise ValueError("Parameter der can only be 1 or 2")
    
    # Compute the magnitude - Does not work for motion blur   
    vals = np.array([sdx.flatten(),sdy.flatten()]).transpose()
    print vals.shape
    stdevs,rotation,center = prcomp(vals,center=True)
    vals = np.dot(vals-center,rotation)
    print rotation
    tmp = np.abs(vals[:,-1]).flatten()
    tmp.sort()
    v = tmp[int(quantile*len(tmp))-1]
    
    if full_output:
        return v,stdevs,rotation,center
    else:
        return v


def sobelDensity(im, normalize = True):
    '''
    This is a measure approximates the edge density measure but the 
    values will be different because different geometric normalzation
    and masking is used.  The "gold standard edge density" has been 
    integrated into the code base for the original csuEvalSystem 2003.
    
    Algorithm by David Bolme
    '''
    mat = im.asMatrix2D()
    
    if normalize:
        mat = pv.meanStd(mat)
    
    # Compute Sobel Filters
    sdx = ndi.sobel(mat,axis=0)
    sdy = ndi.sobel(mat,axis=1)
    
    # Compute the edge magnitudes
    mag = np.sqrt(sdx*sdx + sdy*sdy)
    
    # Return the average
    return mag.mean()


def pixelSymmetry(im, type="SSE", ilog=None):
    '''
    Reflect the Face on the left/right access and compute error.  
    This measures uneven lighting and is an indicator of lighting direction.
    
    type=MSE: Returns Mean Squared Error in the pixel values.
    type=MEAN: Returns mean absolute error.
    type=LR: Returns the Left half mean - Right half mean.  Indicates lighting direction.
    type=TB: Returns the Top half mean - Bottom half mean.  Indicates lighting direction.
    '''
    mat = im.asMatrix2D()
    w,h = mat.shape
    N = w*h
    mat = pv.meanStd(mat)
    reflect = mat[::-1,:]
    diff = (mat-reflect)
    if ilog != None:
        ilog(pv.Image(mat),label="Original")
        ilog(pv.Image(reflect),label="Reflected")
        ilog(pv.Image(diff),label="Difference")
        
    if type == 'MSE':
        return (diff**2).sum()/N
    elif type == 'MEAN':
        return np.abs(diff).sum()/N
    elif type == "LR":
        return mat[:w/2,:].sum()/N - mat[w/2:,:].sum()/N
    elif type == "TB":
        return mat[:,:h/2].sum()/N - mat[:,h/2:].sum()/N
    else:
        raise ValueError("Unknown semetry type: %s"%type)
    
    
    
    
    
    
