# Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
This script copies the images nessisary to run an experiment to a local 
directory and optionally reduces there size.

Created on Jul 19, 2010

@author: bolme
'''

from facerec_util import *

import os
import os.path
import sys
import optparse
import pickle
import shutil

import pyvision as pv
pv.disableCommercialUseWarnings()

import pyvision.analysis.FaceAnalysis.BEE as bee
from pyvision.face.CascadeDetector import CascadeDetector
from pyvision.face.FilterEyeLocator import FilterEyeLocator



def parseOptions():
    '''
    Parse command line options and arguments.
    
    @returns options, args: see the optparse module for details.
    '''
    usage = 'python facerec_getimages.py [OPTIONS] <sigset.xml> <output_dir>'
    parser = optparse.OptionParser(usage=usage)
    
    #parser.add_option('--eyes-xml', dest="eyes_xml", help="add an XML file with eye coordinates", default=[], action='append')
    parser.add_option('--eyes-csv', dest="eyes_csv", help="add a CSV file with eye coordinates", default=[], action='append')
    parser.add_option('--images', dest="image_dir", help="add a directory with images", default=[], action='append')
    parser.add_option('--reduced', dest="reduced", help="use reduced image", default=False, action='store_true')
    parser.add_option('-v','--verbose', dest="verbose", help="print extra information", default=False, action='store_true')
    parser.add_option('--autodetect', dest="autodetect", help="automatically detect faces and eyes", default=False, action='store_true')
    # TODO: add autodetect

    options,args = parser.parse_args()
    if len(args) != 2:
        print "Error: This program expects two arguments on the command line."
        print "usage:", usage
        sys.exit(-1)
    return options, args


def reduceFiles(sigset,image_map,eye_map,options,output_dir):
    '''
    
    '''
    detect_faces = CascadeDetector()
    detect_eyes = FilterEyeLocator()

    i = 0
    for each in sigset:
        subid = each[0]
        recid = each[1][0]['name']
        filename = os.path.basename(each[1][0]['file-name']).split('.')[0]
        basename = parseBasename(image_map[recid])
        outname = os.path.join(output_dir, recid+".jpg")
        if os.path.exists(outname):
            print "The file %s exists.  Skipping..."%outname
            continue


        im = pv.Image(image_map[recid])
        
        if options.autodetect == False:
            leye,reye = eye_map[basename]
        else:
            faces = detect_faces(im)
            if len(faces) < 1:
                print "Warning: No Face Detected for record id: %s    file: %s"%(recid,filename)
                w,h = im.size
                faces = [pv.Rect(0,0,w,h)]
            if len(faces) > 1:
                print "Warning: Multiple Faces Detected for record id: %s    file: %s"%(recid,filename)
            #print dir(faces[0])
            faces.sort(lambda x,y: cmp(-x.area(),-y.area()))
            eyes = detect_eyes(im,faces[:1])

            leye = eyes[0][1]
            reye = eyes[0][2]
            if options.verbose: print "   Eyes Detected at", leye,"and",reye

        affine = pv.AffineFromPoints(leye,reye,REDUCED_LEYE,REDUCED_REYE,REDUCED_SIZE)
        
        tile = affine(im)
        
        outname = os.path.join(output_dir, recid+".jpg")
        print "Saving:",outname
        tile.asPIL().save(outname,quality=95)
        #alg.addTraining(subid,im,None,leye,reye,ilog=None)
        
        i += 1
        if options.verbose: print "Processed training image",i,'of',len(sigset),"  SUB:",subid,"REC:",recid,"FILE:",image_map[recid]
        
    #alg.train(ilog=ilog)

def copyFiles(sigset,image_map,eye_map,options,output_dir):
    '''
    
    '''
    i = 0
    for each in sigset:
        subid = each[0]
        recid = each[1][0]['name']
        filename = os.path.basename(each[1][0]['file-name'])
        basename = parseBasename(image_map[recid])
        outname = os.path.join(output_dir, filename)
        if os.path.exists(outname):
            print "The file %s exists.  Skipping..."%outname
            continue


        shutil.copy(image_map[recid],outname)
        i += 1
        if options.verbose: print "Processed training image",i,'of',len(sigset),"  SUB:",subid,"REC:",recid,"FILE:",image_map[recid]
        
    #alg.train(ilog=ilog)
    


def main():
    options,args = parseOptions()
    
    print "Reading Sigset."
    sigset = bee.parseSigSet(args[0])
    output_dir = args[1]
    
    print "Locating Files."
    image_map = locateFiles(sigset,False,options)
    print "    Found %d of %d files."%(len(image_map),len(sigset))
    assert len(image_map) == len(sigset)
    
    eye_map = {}
    if options.reduced and options.autodetect == False:
        print "Processing Eye Coordinate Data."
        eye_map = readEyeData(image_map,False,options)
        print "    Eyes coordinate determined for %d of %d images."%(len(eye_map),len(image_map))
        assert len(eye_map) == len(image_map)
    elif options.reduced:
        eye_map = {}
        print "Automatically detecting faces and eyes."
        
    print "Copying images."
    if options.reduced:
        reduceFiles(sigset,image_map,eye_map,options,output_dir)
    else:
        copyFiles(sigset,image_map,eye_map,options,output_dir)
        
    print "Copying Complete."


if __name__ == "__main__":
    # If this is the main script, run function main.
    main()
