# Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This script runs a face recognition test on an algorithm.

Created on Jun 29, 2010

@author: bolme
'''
from facerec_util import *

import sys
import optparse
import pickle
import os.path
import traceback

import pyvision as pv
pv.disableCommercialUseWarnings()
import numpy as np

from pyvision.face.CascadeDetector import CascadeDetector
from pyvision.face.FilterEyeLocator import FilterEyeLocator

class ImageLoadError(Exception):
    pass

class FaceDetectionError(Exception):
    pass

class EyeCoordinateError(Exception):
    pass


detect_faces = CascadeDetector()
detect_eyes = FilterEyeLocator()


def parseOptions():
    '''
    Parse command line options and arguments.
    
    @returns options, args: see the optparse module for details.
    '''
    usage = 'python facerec_train.py [OPTIONS] <trained_algorithm.pkl> <query_sigset.xml> <target_sigset.xml> <matrix.mtx>'
    parser = optparse.OptionParser(usage=usage)
    
    #parser.add_option('--eyes-xml', dest="eyes_xml", help="add an XML file with eye coordinates", default=[], action='append')
    parser.add_option('--eyes-csv', dest="eyes_csv", help="add a CSV file with eye coordinates", default=[], action='append')
    parser.add_option('--images', dest="image_dir", help="add a directory with images", default=[], action='append')
    parser.add_option('--reduced', dest="reduced", help="use reduced image", default=False, action='store_true')
    parser.add_option('-v','--verbose', dest="verbose", help="print extra information", default=False, action='store_true')
    #parser.add_option('--autodetect', dest="autodetect", help="automatically detect faces and eyes", default=False, action='store_true')
    parser.add_option('--autodetect', dest="error_log", help="enables automatic face and eye detection and runs in robust mode (any eye coordinates provided will take precedent over automatic detection)", default=None, action='store')
    # TODO: Add auto-detection

    options,args = parser.parse_args()
    if len(args) != 4:
        print "Error: This program expects four arguments on the command line."
        print "usage:", usage
        sys.exit(-1)
    return options, args


def autodetect(im,options,detect_faces,detect_eyes,recid,image_map):
    faces = detect_faces(im)
    if len(faces) < 1:
        raise FaceDetectionError("No Face Detected")
    if len(faces) > 1:
        # dump this to the error log
        options.error_log.write("Warning with recording %s file %s: %s\n" % (recid, image_map[recid], "Multiple Faces Detected."))
        faces.sort(lambda x, y:cmp(-x.area(), -y.area()))
    eyes = detect_eyes(im, faces[:1])
    leye = eyes[0][1]
    reye = eyes[0][2]
    options.error_log.write("Autodetect: Eyes for recording id %s file %s: %f %f %f %f\n" % (recid, image_map[recid], leye.X(),leye.Y(),reye.X(),reye.Y()))

    return leye,reye


def getEyes(alg,eye_map,im,image_map,recid,basename,detect_faces,detect_eyes,options):
    # If auto detecting

    if eye_map.has_key(basename) and eye_map[basename] != None:
        leye, reye = eye_map[basename]   
        
        # determine if the eye coordinates are reversed by compute face angle
        dx = reye.X() - leye.X()
        dy = reye.Y() - leye.Y()
        
        angle = np.arctan2(dy,dx)
        if (2.3561944901923448 < angle or -2.3561944901923448 > angle) and options.error_log:
            # Eye coordinates are probably reversed.
            options.error_log.write("Warning with recording %s file %s: %s\n" % (recid, image_map[recid], "Eye Coordinates May Be Reversed"))
            
        if leye.l2(reye) <= 8.0:
            options.error_log.write("Warning with recording %s file %s: %s\n" % (recid, image_map[recid], "Face is too small"))
        
        w,h = im.size
        if leye.X() < 0 or leye.Y() < 0 or leye.X() > w or leye.Y() > h:
            options.error_log.write("Warning with recording %s file %s: %s\n" % (recid, image_map[recid], "Eye coordinate 1 is outside of image."))

        if reye.X() < 0 or reye.Y() < 0 or reye.X() > w or reye.Y() > h:
            options.error_log.write("Warning with recording %s file %s: %s\n" % (recid, image_map[recid], "Eye coordinate 2 is outside of image."))

        # TODO: Not all algorithms support "preprocess"
        #tile = alg.preprocess(im,None,leye,reye)
        #if isinstance(tile,pv.Image):
        #    w,h = tile.size
        #    if (tile.asMatrix2D() == 0).sum()/float(w*h) > 0.50:
        #        # Face is not in image or image is mostly black, Go to auto detect
        #        options.error_log.write("Warning transformed image is mostly black %s file %s: %s\n" % (recid, image_map[recid], ""))
        #        #return autodetect(im,options,detect_faces,detect_eyes,recid,image_map)
        
        return leye,reye
            
    return autodetect(im,options,detect_faces,detect_eyes,recid,image_map)
    
    


def createFaceRecord(alg, image_map, eye_map, options, detect_faces, detect_eyes, each, recid, basename):
    try:
        subid = each[0]
        recid = each[1][0]['name']
        filename = os.path.basename(each[1][0]['file-name']).split('.')[0]
        basename = parseBasename(image_map[recid])
        if image_map[recid] == None or not os.path.exists(image_map[recid]):
            raise IOError("File does not exist.")
        if not os.path.isfile(image_map[recid]):
            raise IOError("Path is not a normal file.")
        if not os.access(image_map[recid], os.R_OK):
            raise IOError("Could not access file. Check permissions.")
        
        # If the images are corrupted this next line should produce an exception and things fail.
        im = pv.Image(image_map[recid])
        
        
        leye,reye = getEyes(alg,eye_map,im,image_map,recid,basename,detect_faces,detect_eyes,options)
        
        record = alg.getFaceRecord(im, None, leye, reye, ilog=None)
    except Exception as error:
        if options.error_log == None:
            raise
        print "Error with recording %s file %s: %s" % (recid, image_map[recid], error)
        options.error_log.write("Error with recording %s file %s: %s\n" % (recid, image_map[recid], error))
        if options.verbose:
            print traceback.print_exc()
        record = None
    return record

def testAlgorithm(alg,query_sigset, target_sigset,image_map,eye_map,options,ilog=None):
    '''
    After all algorithm tuning data, sigset, image map and eye data has been computed, train the algorithm.
    '''

    queries = []
    i = 0
    for each in query_sigset:
        subid = each[0]
        recid = each[1][0]['name']
        filename = os.path.basename(each[1][0]['file-name']).split('.')[0]
        basename = parseBasename(image_map[recid])
        
        record = createFaceRecord(alg, image_map, eye_map, options, detect_faces, detect_eyes, each, recid, basename)
        
        queries.append(record)

        i += 1
        if options.verbose: print "Processed query image",i,'of',len(query_sigset),"  SUB:",subid,"REC:",recid,"FILE:",image_map[recid]
        
    targets = []
    i = 0
    for each in target_sigset:
        subid = each[0]
        recid = each[1][0]['name']
        filename = os.path.basename(each[1][0]['file-name']).split('.')[0]
        basename = parseBasename(image_map[recid])

        record = createFaceRecord(alg, image_map, eye_map, options, detect_faces, detect_eyes, each, recid, basename)

        targets.append(record)

        i += 1
        if options.verbose: print "Processed target image",i,'of',len(target_sigset),"  SUB:",subid,"REC:",recid,"FILE:",image_map[recid]
        
    allow_missing = options.error_log != None
    if allow_missing:
        raise NotImplementedError("Missing images in dataset -- removed functionality.") # TODO: Reimplement this 
    mat = alg.similarityMatrix(queries,targets)

    return mat

    


def main():
    options,args = parseOptions()
    
    if options.error_log != None:
        options.error_log = open(options.error_log,'w')
    
    print "Loading Algorithm: ",args[0]
    alg = pickle.load(open(args[0],'rb'))
    
    print "Reading Sigsets."
    query_sigset = pv.parseSigSet(args[1])
    target_sigset = pv.parseSigSet(args[2])
    all_sigset = query_sigset+target_sigset
    
    print "Locating %d Files."%len(all_sigset)
    image_map = locateFiles(all_sigset,options.reduced,options)
    
    
    print "Processing Eye Coordinate Data."
    eye_map = readEyeData(image_map,options.reduced,options)
    if options.error_log == None:
        print "    Eyes coordinate determined for %d of %d images."%(len(eye_map),len(image_map))
        assert len(eye_map) == len(image_map)
    
    print "Computing Similarity Matrix."
    mat = testAlgorithm(alg,query_sigset, target_sigset,image_map,eye_map,options)
    
    bee_mat = pv.BEEDistanceMatrix(mat, os.path.basename(args[1]), os.path.basename(args[2]), sigset_dir=os.path.dirname(args[1]), is_distance=False)
    bee_mat.save(args[3])
    
    # ROC data is now produced by facerec_score.py
    roc = bee_mat.getROC()
    print roc.results()



if __name__ == "__main__":
    main()