# Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This script trains a face recognition algorithm

Created on Jun 29, 2010

@author: bolme
'''

from facerec_util import *

import os
import os.path
import sys
import optparse
import pickle

import pyvision as pv
pv.disableCommercialUseWarnings()

from pyvision.face.CascadeDetector import CascadeDetector
from pyvision.face.FilterEyeLocator import FilterEyeLocator

import facerec2010.baseline.lrpca as lrpca
import facerec2010.baseline.kda as kda
import facerec2010.baseline.lrkda as lrkda
import facerec2010.baseline.gaborgrid as gaborgrid
import facerec2010.baseline.tex as tex
import facerec2010.baseline.lda as lda
#import facerec2010.baseline.svmdist as svmdist





def parseOptions():
    '''
    Parse command line options and arguments.
    
    @returns options, args: see the optparse module for details.
    '''
    usage = 'python facerec_train.py [OPTIONS] <training_sigset.xml> <trained_algorithm.pkl>'
    parser = optparse.OptionParser(usage=usage)
    
    #parser.add_option('--eyes-xml', dest="eyes_xml", help="add an XML file with eye coordinates", default=[], action='append')
    parser.add_option('--eyes-csv', dest="eyes_csv", help="add a CSV file with eye coordinates", default=[], action='append')
    parser.add_option('--images', dest="image_dir", help="add a directory with images", default=[], action='append')
    parser.add_option('--alg', dest="alg", help="select an algorithm: PCA, LRPCA, KDA", default='LRPCA', action='store',choices=['PCA','LRPCA','KDA','LRKDA','EBGM','LRPCASVM','TEX','LDA'])
    parser.add_option('--trained-alg', dest="trained_alg", help="A trained algorithm pkl for use with LRPCASVM", default=None, action='store')
    parser.add_option('--tuning', dest="tuning", help="select tuning: GBU, FRGC, FERET, LEFT_OCULAR, RIGHT_OCULAR", default='GBU', action='store',choices=['GBU','FRGC','FERET','LEFT_OCULAR','RIGHT_OCULAR'])
    parser.add_option('--reduced', dest="reduced", help="use reduced image", default=False, action='store_true')
    parser.add_option('-v','--verbose', dest="verbose", help="print extra information", default=False, action='store_true')
    parser.add_option('--autodetect', dest="autodetect", help="automatically detect faces and eyes", default=False, action='store_true')
    parser.add_option('--log-dir', dest="log_dir", help="Create a directory to save images to for debugging purposes", default=None, action='store')
    
    # TODO: add autodetect

    options,args = parser.parse_args()
    if len(args) != 2:
        print "Error: This program expects two arguments on the command line."
        print "usage:", usage
        sys.exit(-1)
    return options, args
    
    
    
def createLRPCA(options):
    '''
    Uses information in the options to create and configure an Local Region
    PCA algorithm.
    '''
    tuning = {}
    
    if options.tuning == "GBU":
        tuning = lrpca.GBU_TUNING
    elif options.tuning == "FRGC":
        tuning = lrpca.GBU_TUNING
    elif options.tuning == "FERET":
        tuning = lrpca.GBU_TUNING
    elif options.tuning == "LEFT_OCULAR":
        tuning = lrpca.LEFT_OCULAR_TUNING
    elif options.tuning == "RIGHT_OCULAR":
        tuning = lrpca.RIGHT_OCULAR_TUNING
    else:
        print "Warning unknown tuning: %s"%options.tuning
        print "    Using default tuning for GBU."
        tuning = lrpca.GBU_TUNING
    
    if options.verbose:
        keys = tuning.keys()
        keys.sort()
        print "Tuning Configuration: %s"%options.tuning
        for key in keys:
            print "   ",key,":",tuning[key]
            
    alg = lrpca.LRPCA(**tuning)
    return alg
        
    
def createTEX(options):
    '''
    Uses information in the options to create and configure an Local Region
    PCA algorithm.
    '''
    tuning = {}
    
    #if options.tuning == "GBU":
    #    tuning = lrpca.GBU_TUNING
    #elif options.tuning == "FRGC":
    #    tuning = lrpca.GBU_TUNING
    #elif options.tuning == "FERET":
    #    tuning = lrpca.GBU_TUNING
    #elif options.tuning == "LEFT_OCULAR":
    #    tuning = lrpca.LEFT_OCULAR_TUNING
    #elif options.tuning == "RIGHT_OCULAR":
    #    tuning = lrpca.RIGHT_OCULAR_TUNING
    #else:
    #    print "Warning unknown tuning: %s"%options.tuning
    #    print "    Using default tuning for GBU."
    #    tuning = lrpca.GBU_TUNING
    
    #if options.verbose:
    #    keys = tuning.keys()
    #    keys.sort()
    #    print "Tuning Configuration: %s"%options.tuning
    #    for key in keys:
    #        print "   ",key,":",tuning[key]
            
    alg = tex.TEX(**tuning)
    return alg

def createLDA(options):
    '''
    Uses information in the options to create and configure an Local Region
    PCA algorithm.
    '''
    tuning = {}
    alg = lda.LDA(**tuning)
    return alg
        
    
def createKDA(options):
    '''
    Uses information in the options to create and configure an Local Region
    PCA algorithm.
    '''
    tuning = {}
    
    if options.tuning == "GBU":
        tuning = kda.GBU_TUNING
    elif options.tuning == "FRGC":
        tuning = kda.GBU_TUNING
    elif options.tuning == "FERET":
        tuning = kda.GBU_TUNING
    else:
        print "Warning unknown tuning: %s"%options.tuning
        print "    Using default tuning for GBU."
        tuning = kda.GBU_TUNING
    
    if options.verbose:
        keys = tuning.keys()
        keys.sort()
        print "Tuning Configuration: %s"%options.tuning
        for key in keys:
            print "   ",key,":",tuning[key]
            
    alg = kda.FaceKDA(**tuning)
    return alg
        
    
def createLRKDA(options):
    '''
    Uses information in the options to create and configure an Local Region
    PCA algorithm.
    '''
    tuning = {}
    
    if options.tuning == "GBU":
        tuning = lrkda.GBU_TUNING
    elif options.tuning == "FRGC":
        tuning = lrkda.GBU_TUNING
    elif options.tuning == "FERET":
        tuning = lrkda.GBU_TUNING
    else:
        print "Warning unknown tuning: %s"%options.tuning
        print "    Using default tuning for GBU."
        tuning = lrkda.GBU_TUNING
    
    if options.verbose:
        keys = tuning.keys()
        keys.sort()
        print "Tuning Configuration: %s"%options.tuning
        for key in keys:
            print "   ",key,":",tuning[key]
            
    alg = lrkda.FaceLRKDA(**tuning)
    return alg
        
    
#def createLRPCASVM(options):
#    '''
#    Uses information in the options to create and configure an Local Region
#    PCA algorithm.
#    '''
#    tuning = {}
#    
#    if options.tuning == "GBU":
#        tuning = lrpca.GBU_TUNING
#    elif options.tuning == "FRGC":
#        tuning = lrpca.GBU_TUNING
#    elif options.tuning == "FERET":
#        tuning = lrpca.GBU_TUNING
#    else:
#        print "Warning unknown tuning: %s"%options.tuning
#        print "    Using default tuning for GBU."
#        tuning = lrpca.GBU_TUNING
#    
#    if options.verbose:
#        keys = tuning.keys()
#        keys.sort()
#        print "Tuning Configuration: %s"%options.tuning
#        for key in keys:
#            print "   ",key,":",tuning[key]
#           
#    if options.trained_alg == None:
#        raise ValueError("--trained-alg option needs to be set on the command line.")
#    
#    alg = pickle.load(open(options.trained_alg,'rb'))
#    
#    return svmdist.SVMDist(alg,svmdist.extract_features,**svmdist.SVM_DIST_LRPCA_CONFIG)
    
        
    
def createGaborGrid(options):
    '''
    Uses information in the options to create and configure an Local Region
    PCA algorithm.
    '''
    tuning = {}
    
    if options.tuning == "GBU":
        tuning = gaborgrid.GBU_TUNING
    elif options.tuning == "FRGC":
        tuning = gaborgrid.GBU_TUNING
    elif options.tuning == "FERET":
        tuning = gaborgrid.GBU_TUNING
    else:
        print "Warning unknown tuning: %s"%options.tuning
        print "    Using default tuning for GBU."
        tuning = gaborgrid.GBU_TUNING
    
    if options.verbose:
        keys = tuning.keys()
        keys.sort()
        print "Tuning Configuration: %s"%options.tuning
        for key in keys:
            print "   ",key,":",tuning[key]
            
    alg = gaborgrid.GaborGrid(**tuning)
    return alg
        
    

    

def trainAlgorithm(alg,sigset,image_map,eye_map,options,ilog=None):
    '''
    After all algorithm tuning data, sigset, image map and eye data has been computed, train the algorithm.
    '''
    detect_faces = CascadeDetector()
    detect_eyes = FilterEyeLocator()

    i = 0
    for each in sigset:
        subid = each[0]
        recid = each[1][0]['name']
        filename = os.path.basename(each[1][0]['file-name']).split('.')[0]
        basename = parseBasename(image_map[recid])

        im = pv.Image(image_map[recid])
        
        if options.autodetect == False:
            leye,reye = eye_map[basename]
        else:
            faces = detect_faces(im)
            if len(faces) < 1:
                print "Warning: No Face Detected for record id: %s    file: %s"%(recid,filename)
                w,h = im.size
                faces = [pv.Rect(0,0,w,h)]
            if len(faces) > 1:
                print "Warning: Multiple Faces Detected for record id: %s    file: %s"%(recid,filename)
            #print dir(faces[0])
            faces.sort(lambda x,y: cmp(-x.area(),-y.area()))
            eyes = detect_eyes(im,faces[:1])

            leye = eyes[0][1]
            reye = eyes[0][2]
            if options.verbose: print "   Eyes Detected at", leye,"and",reye

        alg.addTraining(subid,im,None,leye,reye,ilog=ilog)
        
        i += 1
        if options.verbose: print "Processed training image",i,'of',len(sigset),"  SUB:",subid,"REC:",recid,"FILE:",image_map[recid]
        
    alg.train(ilog=ilog)

        

        

def main():
    options,args = parseOptions()
    
    ilog = None
    
    if options.log_dir != None:
        ilog = pv.ImageLog(topdir = options.log_dir)
    
    print "Reading Sigset."
    sigset = pv.parseSigSet(args[0])
    
    print "Configuring Algorithm."
    alg = None
    if options.alg == "LRPCA":
        alg = createLRPCA(options)
    elif options.alg == "EBGM":
        alg = createGaborGrid(options)
    #elif options.alg == "LRPCASVM":
    #    alg = createLRPCASVM(options)
    elif options.alg == "KDA":
        alg = createKDA(options)
    elif options.alg == "LRKDA":
        alg = createLRKDA(options)
    elif options.alg == "TEX":
        alg = createTEX(options)
    elif options.alg == "LDA":
        alg = createLDA(options)
    else:
        raise ValueError("Error: Cannot create algorithm of type: %s"%options.alg)

    print "Locating Files."
    image_map = locateFiles(sigset,options.reduced,options)
    print "    Found %d of %d files."%(len(image_map),len(sigset))
    assert len(image_map) == len(sigset)
    
    if options.autodetect == False:
        print "Processing Eye Coordinate Data."
        eye_map = readEyeData(image_map,options.reduced,options)
        print "    Eyes coordinate determined for %d of %d images."%(len(eye_map),len(image_map))
        assert len(eye_map) == len(image_map)
    else:
        eye_map = {}
        print "Automatically detecting faces and eyes."
        
    print "Train the algorithm."
    trainAlgorithm(alg,sigset,image_map,eye_map,options,ilog=ilog)
    
    print "Saving Training Output"
    train_file = open(args[1],'wb')
    pickle.dump(alg, train_file)
    print "Training Complete."
    
    if ilog != None:
        ilog.show()


if __name__ == "__main__":
    # If this is the main script, run function main.
    main()
    
    
    
    
    
    
