# Copyright (c) 2010 David S. Bolme
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#  
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#  
#    3. Neither name of copyright holders nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#  
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module contains common utility functions that are used by other 
scripts in this directory.

Created on Jul 16, 2010

@author: bolme
'''

import os
import os.path
import csv

import pyvision as pv

REDUCED_LEYE = pv.Point(128+64,256-20)
REDUCED_REYE = pv.Point(256+128-64,256-20)
REDUCED_SIZE = (512,512)

IMAGE_EXTENSIONS = set(['.TIF','.TIFF','.JPG','.JPEG','.PNG','.GIF','.PPM','.PGM','.BMP'])


def parseBasename(path):
    '''
    This section parses the basename without file extensions from the path.
    
    @returns basename from /path/to/file/basename.eee
    '''
    if path == None:
        return None
    
    dirname,filename = os.path.split(path)
    # Adaptation for FERET sigsets
    basename = filename.split('.')[0]
    #basename,extname = os.path.splitext(filename)
    return basename

def locateFiles(sigset,reduced,options):
    '''
    This function scans the image directories for the files listed in the
    sigset.  If using the reduced image format, the images should be named
    using the recording ID with a standard image extension such as "jpg".
    Otherwise, the directories are scanned for images matching the basename
    of the filename.  This means that the paths in the sigset do not have
    to be specified accuratly but problems do arise if multiple copies
    of an image are in the image directory.
    '''
    image_map = {}
    file_map = {}

    
    for each in sigset:
        rec_id = each[1][0]['name']
        basename = parseBasename(each[1][0]['file-name'])
        image_map[rec_id] = None
        if reduced:
            file_map[rec_id] = rec_id
        else:
            file_map[basename] = rec_id
        
            
    n_images = 0
    
    for imdir in options.image_dir:
        if options.verbose: print "Processing:",imdir
        for rootdir,dirs,files in os.walk(imdir):
            for filename in files:
                basename,ext = os.path.splitext(filename)
                if ext.upper() not in IMAGE_EXTENSIONS:
                    continue
                if file_map.has_key(basename) and image_map[file_map[basename]] == None:
                    image_map[file_map[basename]] = os.path.join(rootdir,filename)
                    if options.verbose: print "    Found image:",basename,image_map[file_map[basename]]
                    n_images += 1
                elif file_map.has_key(basename) and image_map[file_map[basename]] != None:
                    raise ValueError("Multiple images found matching recording id %s:\n   First instance:  %s\n   Second instance: %s"%(file_map[basename],image_map[file_map[basename]],os.path.join(rootdir,filename)))
                    
                    
    if options.verbose: print "Found %d of %d images."%(n_images,len(image_map))
    
    if n_images != len(image_map):
        if hasattr(options,'error_log') and options.error_log != None:
            options.error_log.write("Missing %d images.\n"%(len(image_map)-n_images,))
            print "Error: Missing %d images."%(len(image_map)-n_images,)
        else:
            raise ValueError("Missing %d images."%(len(image_map)-n_images,)) # TODO: make a better exception class
    
    return image_map
    

def readEyeData(image_map,reduced,options):
    '''
    Read eye data files and match eye coordinates to recording ids.
    '''
    eye_map = {}
    
    for key in image_map.keys():
        basename = parseBasename(image_map[key])
        eye_map[basename] = None

    if reduced:
        # In reduced image format eyes are always in the same location.
        for key in image_map.keys():
            eye_map[parseBasename(image_map[key])] = [REDUCED_LEYE, REDUCED_REYE]        
        
    for eyefile in options.eyes_csv:
        f = csv.reader(open(eyefile,'rb'))
        for row in f:
            try:
                fname = parseBasename(row[0])
                leyex = float(row[1])
                leyey = float(row[2])
                reyex = float(row[3])
                reyey = float(row[4])
                if eye_map.has_key(fname):
                    eye_map[fname] = (pv.Point(leyex,leyey),pv.Point(reyex,reyey))
            except:
                print "Could not parse row:",row
                
    for key in eye_map.keys():
        if eye_map[key] == None:
            del eye_map[key] 
    
    if options.verbose:
        print "Eye coordinate data:"
        for key in image_map.keys():
            basename = parseBasename(image_map[key])
            if eye_map.has_key(basename):
                print "    %s : %s"%(key,eye_map[basename])
            else:
                print "    %s : Automatic"%(basename,)
    return eye_map
    
